#
# tree.rb
#
#   Copyright (c) 2001 Minero Aoki <aamine@cd.xdsl.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU General Public License version 2 or later.
#

require 'amstd/bug'


module EWB

  class ParseError < StandardError; end


  module Acceptable

    include Enumerable

    def accept( v )
      v.__send__( "ewb_#@name", self ) do
        @children.each do |c|
          c.accept v
        end
      end
    end

    def each( &block )
      @children.each &block
    end
  
  end


  class Toplevel
  
    include Acceptable

    def initialize
      @name = 'toplevel'
      @children = []
    end

    def level
      -1
    end

    def child( block )
      block.name == 'i' or raise ParseError, 'block must be //i for toplevel'
      @children.push block
    end

    def close
      raise ParseError, "too many block close: //#{name}}"
    end
    
    def leave
    end

    def inspect
      '#<ewb.toplevel>'
    end
  
  end


  class Clause

    include Acceptable

    def initialize( name, no )
      @name = name
      @lineno = no

      @children = []
      @closed = false
    end

    attr :name
    attr :lineno
  
    def child( c )
      @children.push c
    end

    def close( name )
      if name then
        name == @name or
            raise ParseError, "wrong block close //#{name}} at #{@lineno}"
      end
      @closed = true
    end

    def leave
      @closed or raise ParseError, "#{@name} block is not closed"
    end

    def display
      puts "//#@name begin"
      @children.each do |c|
        c.display
      end
      puts "//#@name end"
    end

    def inspect
      "#<ewb.#@name>"
    end

  end


  class Block < Clause

    def self.create( name, arg, label, no )
      lev, klass = DESCRIPTER[name]
      lev or raise ParseError, "unknown block trigger: #{name}"
      klass.new lev, name, arg, label, no
    end

    def initialize( lev, name, arg, label, no )
      super name, no
      @level = lev
      @args = arg
      @label = label
    end

    attr :level
    attr :args
    attr :label

    def toplevel?
      @level == 1
    end

    def inspect
      "\#<ewb.#@name level=#@level>"
    end

  end


  class HeadBlock < Block

    def close( nm )
      raise ParseError, "too many block close: //#{nm.to_s}} at #@lineno"
    end

    def leave
    end

  end


  class Block
    DESCRIPTER = {
      'i'      => [ 0, HeadBlock ],
      'ii'     => [ 1, HeadBlock ],
      'iii'    => [ 2, HeadBlock ],
      'iiii'   => [ 3, HeadBlock ],
      'iiiii'  => [ 4, HeadBlock ],
      'iiiiii' => [ 5, HeadBlock ],

      'note'    => [ 10, Block ],
      'warning' => [ 10, Block ],
      'ky'      => [ 10, Block ],
      'read'    => [ 10, Block ],

      'lst1'    => [ 100, Block ],
      'lst2'    => [ 100, Block ],
      'k1'      => [ 100, Block ],
      'k2'      => [ 100, Block ],
      'tabs'    => [ 100, Block ],
      'tabm'    => [ 100, Block ],
      'tabl'    => [ 100, Block ],
      'cb'      => [ 100, Block ]
    }

    NAMES = []
    DESCRIPTER.keys.each do |i|
      NAMES.push i unless /\Ai+\z/ === i
    end
  end


  class Paragraph < Clause

    def initialize( no )
      super 'paragraph', no
    end

    def close( name )
      raise ParseError, "too many block close: //#{name.to_s}} at #{lineno}"
    end

    def display
      puts "//paragraph begin"
      @children.each do |c|
        c.display 1
      end
      puts "//paragraph end"
    end

    def empty?
      @children.size == 1 and
      Text === @children[0] and
      @children[0].empty?
    end

  end


  class InlineBlock < Clause

    def display( n = 0 )
      print '  ' * n
      puts name
      @children.each do |c|
        c.display n + 1
      end
    end
  
  end


  class InlineSingle

    def initialize( name, no )
      @name = name
      @lineno = no
    end

    attr :name
    attr :lineno

    def accept( v )
      v.__send__ "ewb_#{@name}", self
    end

    def each
    end

    def display( n = 0 )
      print '  ' * n
      puts name
    end

    def inspect
      "\#<ewb.#{@name}>"
    end
  
  end


  class EntityDef < InlineBlock

    def initialize( name, num, no )
      super name + '_def', no
      @number = num
    end

    attr :number

    alias ident number
  
  end


  class EntityRef < InlineSingle

    def initialize( name, num, no )
      super name + '_ref', no
      @number = num
    end

    attr :number

    alias ident number

  end


  class Text

    def initialize( str, no )
      @text = str
      @lineno = no
    end

    attr :text
    attr :lineno

    def accept( v )
      v.ewb_text @text
    end

    def each
    end

    def empty?
      @text.strip.empty?
    end
  
    def display( n = 0 )
      print '  ' * n
      p @text
    end

    def inspect
      "\#<ewb.#@name>"
    end

  end


  module Parser

    names = Block::NAMES.join('|')
    BLOCK_BEGIN = %r~\A//(#{names})(\[[^\]]+\])?\{~


    module_function

    def parse( f )
      stack = []
      stack.push toplevel = Toplevel.new

      f.each do |line|
        case line
        when %r<\A//(i+)>, BLOCK_BEGIN
          tmp = Block.create( $1, $2, $'.strip, f.lineno )
          until tmp.level > stack.last.level do
            stack.pop.leave
#print ")#{f.lineno}"; $stdout.flush
#puts if stack.size == 1
          end
          stack.last.child tmp
          stack.push tmp
#print "(#{f.lineno}"; $stdout.flush

        when %r<\A//(\w+)?\}>
          stack.last.close $1
          stack.pop.leave
#print ")#{f.lineno}"; $stdout.flush

        else
begin
          stack.last.child parse_paragraph( line, f.lineno )
rescue ParseError
p self
raise
end
        end
      end
      until stack.empty? do
        stack.pop.leave
      end

      toplevel
    end


    WITH_N = %w(
      ky f l t
    )

    withn = WITH_N.join('|')
    WITH_N_DEF = %r~\A//(#{withn})(\d+) ~
    WITH_N_REF = %r~\A//(#{withn})(\d+)~

    def parse_paragraph( str, lineno )
      para = Paragraph.new(lineno)
      stack = [ para ]

      until str.empty? do
        if m = %r~//(?:[a-z_0-9]+\{|[a-z_0-9]*\}|[a-z_0-9]+ ?|//)~i.match(str) then
          str = m.post_match
          text m.pre_match, stack

          case m[0]
          when %r~\A//([a-z_0-9]+)\{~i
            i = InlineBlock.new( $1, lineno )
            stack.last.child i
            stack.push i

          when %r~\A//([a-z_0-9]+)?\}~i
            stack.pop.close $1

          when WITH_N_DEF
            stack.size == 1 or
                    raise ParseError, 'entity define must be placed on begging'
            c = EntityDef.new( $1, $2, lineno )
            stack = [c]
            
          when WITH_N_REF
            stack.last.child EntityRef.new( $1, $2, lineno )

          when %r~\A//([a-z_0-9]+)~i
            stack.last.child InlineSingle.new( $1, lineno )

          when %r~\A////~
            text '//', stack

          else
            bug! 'inline parse'
          end
        
        else
          text str, stack
          break
        end
      end
      stack.size == 1 or
              raise ParseError, "unclosed inline block at #{lineno}"

      stack[0]   ##### tmp
    end

    def text( str, stack )
      stack.last.child Text.new(str, stack[0].lineno) unless str.empty?
    end
  
  end

end
